//===--- PPCallbacks.cpp - Callbacks for Preprocessor actions ---*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//

#include "ModularizePPCallbacks.h"
#include "ModularizeUtilities.h"
#include "llvm/Support/raw_ostream.h"
#include "clang/Basic/SourceManager.h"
#include <ctype.h>
#include <stdarg.h>

namespace Modularize {
using namespace llvm;
using namespace clang;

// FIXME: This class doesn't need to override all the handlers.

// Set to 1 to enable tracing messages.
// FIXME: Take out the tracing when it will no longer be needed.
#define MPPC_TRACE 0

ModularizePPCallbacks::ModularizePPCallbacks(Preprocessor& pp, StringRef topFile,
      ModularizeMasterHeaderTracker &masterHeaderTracker) :
    MasterHeaderTracker(masterHeaderTracker),
  TopFile(topFile.str()),
    PP(pp),
    RootHeaderFile(NULL),
    CurrentHeaderFile(NULL) {
  ModularizeHeaderFile *mhf = new ModularizeHeaderFile(topFile.str(), PP);
  AddHeaderFile(mhf);
  CurrentHeaderFile = mhf;
  RootHeaderFile = mhf;
}

ModularizePPCallbacks::~ModularizePPCallbacks() {
#if MPPC_TRACE
  Trace("~ModularizePPCallbacks called.\n");
#endif
}

/// \brief Callback invoked whenever a source file is entered or exited.
///
/// \param Loc Indicates the new location.
/// \param PrevFID the file that was exited if \p Reason is ExitFile.
void ModularizePPCallbacks::FileChanged(SourceLocation Loc, FileChangeReason Reason,
                         SrcMgr::CharacteristicKind FileType,
                         FileID PrevFID) {
  std::string locFileName(SourceLocationFile(Loc));
#if MPPC_TRACE
  std::string prevFileName(FileName(PrevFID));
  std::string locString(SourceLocationStr(Loc));
  Trace(Loc, "FileChanged(%s, %s, %s, %s) called.\n",
    locString.c_str(), FileChangeReasonStr(Reason), CharacteristicKindStr(FileType),
    prevFileName.c_str());
#endif
  ModularizeHeaderFile *mhf = NULL;
  std::string *fileName = NULL;
  switch (Reason) {
    case EnterFile:
      fileName = &locFileName;
      break;
    case ExitFile:
      if (PrevFID.isInvalid())
        fileName = &TopFile;
      else
        fileName = &locFileName;
      break;
    case SystemHeaderPragma:
      return;
    case RenameFile:
      return;
    default:
      return;
  }
  if (fileName == NULL)
    return;
  mhf = GetHeaderFile(*fileName);
  if (mhf == NULL) {
    mhf = new ModularizeHeaderFile(*fileName, PP);
    AddHeaderFile(mhf);
    CurrentHeaderFile = mhf;
    if (!RootHeaderFile)
      RootHeaderFile = mhf;
  }
  else
    CurrentHeaderFile = mhf;
}

/// \brief Callback invoked whenever a source file is skipped as the result
/// of header guard optimization.
///
/// \param ParentFile The file that \#included the skipped file.
///
/// \param FilenameTok The token in ParentFile that indicates the
/// skipped file.
void ModularizePPCallbacks::FileSkipped(const FileEntry &ParentFile,
                         const Token &FilenameTok,
                         SrcMgr::CharacteristicKind FileType) {
#if MPPC_TRACE
  Trace(FilenameTok, "FileSkipped called.\n");
#endif
}

/// \brief Callback invoked whenever an inclusion directive results in a
/// file-not-found error.
///
/// \param FileName The name of the file being included, as written in the 
/// source code.
///
/// \param RecoveryPath If this client indicates that it can recover from 
/// this missing file, the client should set this as an additional header
/// search patch.
///
/// \returns true to indicate that the preprocessor should attempt to recover
/// by adding \p RecoveryPath as a header search path.
bool ModularizePPCallbacks::FileNotFound(StringRef FileName,
                          SmallVectorImpl<char> &RecoveryPath) {
#if MPPC_TRACE
  Trace("FileNotFound(%s) called.\n", FileName.str().c_str());
#endif
  return false;
}

/// \brief Callback invoked whenever an inclusion directive of
/// any kind (\c \#include, \c \#import, etc.) has been processed, regardless
/// of whether the inclusion will actually result in an inclusion.
///
/// \param HashLoc The location of the '#' that starts the inclusion 
/// directive.
///
/// \param IncludeTok The token that indicates the kind of inclusion 
/// directive, e.g., 'include' or 'import'.
///
/// \param FileName The name of the file being included, as written in the 
/// source code.
///
/// \param IsAngled Whether the file name was enclosed in angle brackets;
/// otherwise, it was enclosed in quotes.
///
/// \param FilenameRange The character range of the quotes or angle brackets
/// for the written file name.
///
/// \param File The actual file that may be included by this inclusion 
/// directive.
///
/// \param SearchPath Contains the search path which was used to find the file
/// in the file system. If the file was found via an absolute include path,
/// SearchPath will be empty. For framework includes, the SearchPath and
/// RelativePath will be split up. For example, if an include of "Some/Some.h"
/// is found via the framework path
/// "path/to/Frameworks/Some.framework/Headers/Some.h", SearchPath will be
/// "path/to/Frameworks/Some.framework/Headers" and RelativePath will be
/// "Some.h".
///
/// \param RelativePath The path relative to SearchPath, at which the include
/// file was found. This is equal to FileName except for framework includes.
///
/// \param Imported The module, whenever an inclusion directive was
/// automatically turned into a module import or null otherwise.
///
void ModularizePPCallbacks::InclusionDirective(SourceLocation HashLoc,
                                const Token &IncludeTok,
                                StringRef FileName,
                                bool IsAngled,
                                CharSourceRange FilenameRange,
                                const FileEntry *File,
                                StringRef SearchPath,
                                StringRef RelativePath,
                                const Module *Imported) {
  std::string fileName(File->getName());
#if 0 // Can result in mismatched files.
  ModularizeHeaderFile *mhf = NULL;
  mhf = GetHeaderFile(fileName);
  if (mhf == NULL) {
    mhf = new ModularizeHeaderFile(fileName, PP);
    AddHeaderFile(mhf);
    CurrentHeaderFile = mhf;
    if (!RootHeaderFile)
      RootHeaderFile = mhf;
  }
  else
    CurrentHeaderFile = mhf;
#endif
#if MPPC_TRACE
  Trace(IncludeTok, "InclusionDirective(%s) called.\n", fileName.c_str());
#endif
}

/// \brief Callback invoked whenever there was an explicit module-import
/// syntax.
///
/// \param ImportLoc The location of import directive token.
///
/// \param Path The identifiers (and their locations) of the module
/// "path", e.g., "std.vector" would be split into "std" and "vector".
///
/// \param Imported The imported module; can be null if importing failed.
///
void ModularizePPCallbacks::moduleImport(SourceLocation ImportLoc,
                          ModuleIdPath Path,
                          const Module *Imported) {
#if MPPC_TRACE
  Trace(ImportLoc, "moduleImport() called.\n");
#endif
}

/// \brief Callback invoked when the end of the main file is reached.
///
/// No subsequent callbacks will be made.
void ModularizePPCallbacks::EndOfMainFile() {
  // Add all the header snapshots to the master header file tracker.
  ModularizeHeaderFileMapIterator iter = HeaderFileMap.begin();
  ModularizeHeaderFileMapIterator end = HeaderFileMap.end();
  for (; iter != end; ++iter) {
    MasterHeaderTracker.AddHeaderFile((*iter).second, TopFile);
  }
#if MPPC_TRACE
  Trace("EndOfMainFile() called.\n");
#endif
}

/// \brief Callback invoked when a \#ident or \#sccs directive is read.
/// \param Loc The location of the directive.
/// \param str The text of the directive.
///
void ModularizePPCallbacks::Ident(SourceLocation Loc, const std::string &str) {
#if MPPC_TRACE
  Trace(Loc, "Ident() called.\n");
#endif
}

/// \brief Callback invoked when a \#pragma comment directive is read.
void ModularizePPCallbacks::PragmaComment(SourceLocation Loc, const IdentifierInfo *Kind,
                           const std::string &Str) {
#if MPPC_TRACE
  Trace(Loc, "PragmaComment() called.\n");
#endif
}

/// \brief Callback invoked when a \#pragma clang __debug directive is read.
/// \param Loc The location of the debug directive.
/// \param DebugType The identifier following __debug.
void ModularizePPCallbacks::PragmaDebug(SourceLocation Loc, StringRef DebugType) {
#if MPPC_TRACE
  Trace(Loc, "PragmaDebug() called.\n");
#endif
}

/// \brief Callback invoked when a \#pragma message directive is read.
/// \param Loc The location of the message directive.
/// \param Namespace The namespace of the message directive.
/// \param Kind The type of the message directive.
/// \param Str The text of the message directive.
void ModularizePPCallbacks::PragmaMessage(SourceLocation Loc, StringRef Namespace,
                           PragmaMessageKind Kind, StringRef Str) {
#if MPPC_TRACE
  Trace(Loc, "PragmaMessage() called.\n");
#endif
}

/// \brief Callback invoked when a \#pragma gcc dianostic push directive
/// is read.
void ModularizePPCallbacks::PragmaDiagnosticPush(SourceLocation Loc,
                                  StringRef Namespace) {
#if MPPC_TRACE
  Trace(Loc, "PragmaDiagnosticPush() called.\n");
#endif
}

/// \brief Callback invoked when a \#pragma gcc dianostic pop directive
/// is read.
void ModularizePPCallbacks::PragmaDiagnosticPop(SourceLocation Loc,
                                 StringRef Namespace) {
#if MPPC_TRACE
  Trace(Loc, "PragmaDiagnosticPop() called.\n");
#endif
}

/// \brief Callback invoked when a \#pragma gcc dianostic directive is read.
void ModularizePPCallbacks::PragmaDiagnostic(SourceLocation Loc, StringRef Namespace,
                              diag::Mapping mapping, StringRef Str) {
#if MPPC_TRACE
  Trace(Loc, "PragmaDiagnostic() called.\n");
#endif
}

/// \brief Called by Preprocessor::HandleMacroExpandedIdentifier when a
/// macro invocation is found.
void ModularizePPCallbacks::MacroExpands(const Token &MacroNameTok, const MacroDirective *MD,
                          SourceRange Range, const MacroArgs *Args) {
  std::string macroName = MacroNameTok.getIdentifierInfo()->getName().str();
  std::string value;
  const MacroInfo *MI = PP.getMacroInfo(MacroNameTok.getIdentifierInfo());
  MacroInfo::tokens_iterator I = MI->tokens_begin();
  MacroInfo::tokens_iterator E = MI->tokens_end();
  for (; I != E; ++I)
    value += PP.getSpelling((*I));
  AddSymbol(macroName, value);
#if MPPC_TRACE
  Trace(MacroNameTok, "MacroExpands(%s) called: ", macroName.c_str());
  for (I = MI->tokens_begin(); I != E; ++I) {
    Trace(" '%s'", PP.getSpelling((*I)).c_str());
  }
  Trace("\n");
#endif
}

/// \brief Hook called whenever a macro definition is seen.
void ModularizePPCallbacks::MacroDefined(const Token &MacroNameTok,
                          const MacroDirective *MD) {
#if MPPC_TRACE
  std::string macroName =
    MacroNameTok.getIdentifierInfo()->getName().str();
  Trace(MacroNameTok, "MacroDefined(%s) called: ", macroName.c_str());
  const MacroInfo *MI = PP.getMacroInfo(MacroNameTok.getIdentifierInfo());
  MacroInfo::tokens_iterator I = MI->tokens_begin();
  MacroInfo::tokens_iterator E = MI->tokens_end();
  for (; I != E; ++I) {
    Trace(" '%s'", PP.getSpelling((*I)).c_str());
  }
  Trace("\n");
#endif
}

/// \brief Hook called whenever a macro \#undef is seen.
///
/// MD is released immediately following this callback.
void ModularizePPCallbacks::MacroUndefined(const Token &MacroNameTok,
                            const MacroDirective *MD) {
#if MPPC_TRACE
  std::string macroName =
    MacroNameTok.getIdentifierInfo()->getName().str();
  Trace(MacroNameTok, "MacroUndefined(%s) called.\n", macroName.c_str());
#endif
}

/// \brief Hook called whenever the 'defined' operator is seen.
/// \param MD The MacroDirective if the name was a macro, null otherwise.
void ModularizePPCallbacks::Defined(const Token &MacroNameTok, const MacroDirective *MD) {
#if MPPC_TRACE
  std::string macroName =
    MacroNameTok.getIdentifierInfo()->getName().str();
  Trace(MacroNameTok, "Defined(%s) called.\n", macroName.c_str());
#endif
}

/// \brief Hook called when a source range is skipped.
/// \param Range The SourceRange that was skipped. The range begins at the
/// \#if/\#else directive and ends after the \#endif/\#else directive.
void ModularizePPCallbacks::SourceRangeSkipped(SourceRange Range) {
#if MPPC_TRACE
  Trace(Range.getBegin(), "SourceRangeSkipped() called.\n");
#endif
}

/// \brief Hook called whenever an \#if is seen.
/// \param Loc the source location of the directive.
/// \param ConditionRange The SourceRange of the expression being tested.
///
// FIXME: better to pass in a list (or tree!) of Tokens.
void ModularizePPCallbacks::If(SourceLocation Loc, SourceRange ConditionRange) {
  std::string unconvertedConditional(SourceSnippet(ConditionRange));
  std::string convertedConditional(DoMacroSubstitutions(unconvertedConditional));
  unsigned lineNumber = SourceLocationLineNumber(Loc);
  if (CurrentHeaderFile->GetLastLineNumber() < lineNumber) {
    ModularizePPDirective directive(
      MPPD_If, lineNumber,
      unconvertedConditional, convertedConditional);
    CurrentHeaderFile->AddDirective(directive);
  }
#if MPPC_TRACE
  Trace(Loc, "If(%s (%s)) called.\n",
    unconvertedConditional.c_str(), convertedConditional.c_str());
  Trace("CurrentHeaderFile.Name = %s\n", CurrentHeaderFile->GetName().c_str());
#endif
}

/// \brief Hook called whenever an \#elif is seen.
/// \param Loc the source location of the directive.
/// \param ConditionRange The SourceRange of the expression being tested.
/// \param IfLoc the source location of the \#if/\#ifdef/\#ifndef directive.
// FIXME: better to pass in a list (or tree!) of Tokens.
void ModularizePPCallbacks::Elif(SourceLocation Loc, SourceRange ConditionRange,
                  SourceLocation IfLoc) {
  std::string unconvertedConditional(SourceSnippet(ConditionRange));
  std::string convertedConditional(DoMacroSubstitutions(unconvertedConditional));
  unsigned lineNumber = SourceLocationLineNumber(Loc);
  if (CurrentHeaderFile->GetLastLineNumber() < lineNumber) {
    ModularizePPDirective directive(
      MPPD_ElIf, SourceLocationLineNumber(Loc),
      unconvertedConditional, convertedConditional);
    CurrentHeaderFile->AddDirective(directive);
  }
#if MPPC_TRACE
  Trace(Loc, "Elif(%s (%s)) called.\n",
    unconvertedConditional.c_str(), convertedConditional.c_str());
  Trace("CurrentHeaderFile.Name = %s\n", CurrentHeaderFile->GetName().c_str());
#endif
}

/// \brief Hook called whenever an \#ifdef is seen.
/// \param Loc the source location of the directive.
/// \param MacroNameTok Information on the token being tested.
/// \param MD The MacroDirective if the name was a macro, null otherwise.
void ModularizePPCallbacks::Ifdef(SourceLocation Loc, const Token &MacroNameTok,
                   const MacroDirective *MD) {
  std::string macroName = MacroNameTok.getIdentifierInfo()->getName().str();
  std::string unconvertedConditional(macroName);
  std::string convertedConditional(
    macroName + (MD ? " (defined)" : " (not defined)"));
  unsigned lineNumber = SourceLocationLineNumber(Loc);
  if (CurrentHeaderFile->GetLastLineNumber() < lineNumber) {
    ModularizePPDirective directive(
      MPPD_IfDef, SourceLocationLineNumber(Loc),
      unconvertedConditional, convertedConditional);
    CurrentHeaderFile->AddDirective(directive);
  }
#if MPPC_TRACE
  Trace(Loc, "Ifdef(%s, %s) called.\n",
    macroName.c_str(), (MD ? " (defined)" : " (not defined)"));
  Trace("CurrentHeaderFile.Name = %s\n", CurrentHeaderFile->GetName().c_str());
#endif
}

/// \brief Hook called whenever an \#ifndef is seen.
/// \param Loc the source location of the directive.
/// \param MacroNameTok Information on the token being tested.
/// \param MD The MacroDirective if the name was a macro, null otherwise.
void ModularizePPCallbacks::Ifndef(SourceLocation Loc, const Token &MacroNameTok,
                    const MacroDirective *MD) {
  std::string macroName = MacroNameTok.getIdentifierInfo()->getName().str();
  std::string unconvertedConditional(macroName);
  std::string convertedConditional(
    macroName + (MD ? " (defined)" : " (not defined)"));
  unsigned lineNumber = SourceLocationLineNumber(Loc);
  if (CurrentHeaderFile->GetLastLineNumber() < lineNumber) {
    ModularizePPDirective directive(
      MPPD_IfNDef, SourceLocationLineNumber(Loc),
      unconvertedConditional, convertedConditional);
    CurrentHeaderFile->AddDirective(directive);
  }
#if MPPC_TRACE
  Trace(Loc, "Ifndef(%s, %s) called.\n",
    macroName.c_str(), (MD ? " (defined)" : " (not defined)"));
  Trace("CurrentHeaderFile.Name = %s\n", CurrentHeaderFile->GetName().c_str());
#endif
}

/// \brief Hook called whenever an \#else is seen.
/// \param Loc the source location of the directive.
/// \param IfLoc the source location of the \#if/\#ifdef/\#ifndef directive.
void ModularizePPCallbacks::Else(SourceLocation Loc, SourceLocation IfLoc) {
#if MPPC_TRACE
  Trace(Loc, "Else() called.\n");
#endif
}

/// \brief Hook called whenever an \#endif is seen.
/// \param Loc the source location of the directive.
/// \param IfLoc the source location of the \#if/\#ifdef/\#ifndef directive.
void ModularizePPCallbacks::Endif(SourceLocation Loc, SourceLocation IfLoc) {
#if MPPC_TRACE
  Trace(Loc, "Endif() called.\n");
#endif
}

// Some utility functions.

std::string ModularizePPCallbacks::SourceLocationStr(SourceLocation Loc) {
  if (Loc.isInvalid())
    return std::string("(none)");
  else
    return Loc.printToString(PP.getSourceManager());
}

std::string ModularizePPCallbacks::SourceLocationFile(SourceLocation Loc) {
  std::string source(SourceLocationStr(Loc));
  int offset = source.find(':', 2);
  if (offset < 0)
    return source;
  return source.substr(0, offset);
}

unsigned ModularizePPCallbacks::SourceLocationLineNumber(SourceLocation Loc) {
  return PP.getSourceManager().getPresumedLineNumber(Loc);
}

std::string ModularizePPCallbacks::FileName(FileID fileID) {
  return SourceLocationFile(PP.getSourceManager().getLocForStartOfFile(fileID));
}

ModularizeHeaderFile *ModularizePPCallbacks::GetHeaderFile(const std::string& name) {
  ModularizeHeaderFileMap::iterator iter = HeaderFileMap.find(name);
  if (iter != HeaderFileMap.end())
    return iter->second;
  return NULL;
}

void ModularizePPCallbacks::AddHeaderFile(ModularizeHeaderFile *headerFile) {
  HeaderFileMap[headerFile->GetName()] = headerFile;
}

const char *ModularizePPCallbacks::FileChangeReasonStr(FileChangeReason reason) {
  switch (reason) {
  case EnterFile:
    return "EnterFile";
  case ExitFile:
    return "ExitFile";
  case SystemHeaderPragma:
    return "SystemHeaderPragma";
  case RenameFile:
    return "RenameFile";
  default:
    return "(unknown)";
  }
}

const char *ModularizePPCallbacks::CharacteristicKindStr(SrcMgr::CharacteristicKind fileType) {
  using namespace SrcMgr;
  switch (fileType) {
  case C_User:
    return "C_User";
  case C_System:
    return "C_System";
  case C_ExternCSystem:
    return "C_ExternCSystem";
  default:
    return "(unknown)";
  }
}

// Retrieve source snippet from file image.
std::string ModularizePPCallbacks::SourceSnippet(SourceRange sourceRange) {
  SourceLocation bLoc(sourceRange.getBegin());
  SourceLocation eLoc(sourceRange.getEnd());

  // Decompose the locations into FID/Offset pairs.
  std::pair<FileID, unsigned> bLocInfo = PP.getSourceManager().getDecomposedLoc(bLoc);
  std::pair<FileID, unsigned> eLocInfo = PP.getSourceManager().getDecomposedLoc(eLoc);
  FileID FID = bLocInfo.first;
  unsigned bFileOffset = bLocInfo.second;
  unsigned eFileOffset = eLocInfo.second;
  unsigned length = eFileOffset - bFileOffset;

  // Get information about the buffer it points into.
  bool Invalid = false;
  const char *BufStart = PP.getSourceManager().getBufferData(FID, &Invalid).data();
  if (Invalid)
    return std::string();

  // Rewind from the current position to the start of the line.
  const char *bPtr = BufStart + bFileOffset;

  // Trim snippet.
  while ((*bPtr <= ' ') && (length != 0)) {
    bPtr++;
    length--;
  }

  while ((length != 0) && (bPtr[length - 1] <= ' '))
    length--;

  std::string returnValue(bPtr, length);

  return returnValue;
}

// Do macro substitutions in a string.
// FIXME: Note that this is very simple parser, such that it doesn't
// support things like function-style macros, macros with string
// or character literals, and defined().
// FIXME: There probably should be a function in the Preprocessor
// class that will lex a string, do substitutions, and return either
// a token list or a converted string.
std::string ModularizePPCallbacks::DoMacroSubstitutions(std::string str) {
  enum LexMode {
    LM_Top,
    LM_Symbol
  } lexMode = LM_Top;
  char output[1024];
  int outputLength = 0;
  char token[256];
  int tokenLength = 0;
  std::string symbolValue;
  int symbolLength;
  int length = str.length();
  int index;
  char c = '\0';
  for (index = 0; index <= length; index++) {
    if (index < length)
      c = str[index];
    else
      c = '\0';
restart:
    switch (lexMode) {
      case LM_Top:
        tokenLength = 0;
        if (isalpha(c) || (c == '_')) {
          lexMode = LM_Symbol;
          goto restart;
        }
        else {
          if (outputLength < sizeof(output) + 1)
            output[outputLength++] = c;
        }
        break;
      case LM_Symbol:
        if (isalpha(c) || (c == '_') || isdigit(c)) {
          if (tokenLength < sizeof(token) - 1)
            token[tokenLength++] = c;
        }
        else {
          token[tokenLength] = '\0';
          if (GetSymbol(token, symbolValue)) {
            // Recursively do substitutions on macro value.
            symbolValue = DoMacroSubstitutions(symbolValue);
            symbolLength = symbolValue.length();
            if (outputLength + symbolLength < sizeof(output) - 1) {
              strcpy(output + outputLength, symbolValue.c_str());
              outputLength += symbolLength;
            }
            else {
              strncpy(output + outputLength, symbolValue.c_str(), sizeof(output) - outputLength - 1);
              outputLength = sizeof(output) - 1;
              output[sizeof(output) - 1] = '\0';
            }
          }
          else {
            if (outputLength + tokenLength < sizeof(output) - 1) {
              strcpy(output + outputLength, token);
              outputLength += tokenLength;
            }
            else {
              strncpy(output + outputLength, token, sizeof(output) - outputLength - 1);
              outputLength = sizeof(output) - 1;
              output[sizeof(output) - 1] = '\0';
            }
          }
          lexMode = LM_Top;
          goto restart;
        }
        break;
    }
  }
  output[outputLength] = '\0';
  return std::string(output);
}

// Add a symbol definition.  If the symbol already exists,
// its value is replaced.
void ModularizePPCallbacks::AddSymbol(std::string name, std::string value) {
  Symbols[name] = value;
}

// Get symbol value.  Return empty string if not defined.
std::string ModularizePPCallbacks::GetSymbol(std::string name) {
  ModularizeSymbolMapIterator iter = Symbols.find(name);
  if (iter != Symbols.end())
    return (*iter).second;
  return std::string();
}

// Get symbol value.  Return true if found.
bool ModularizePPCallbacks::GetSymbol(std::string name, std::string &value) {
  ModularizeSymbolMapIterator iter = Symbols.find(name);
  if (iter != Symbols.end()) {
    value = (*iter).second;
  return true;
  }
  return false;
}

void ModularizePPCallbacks::Trace(const char *format, ...) {
  char buffer[1024];
	va_list	args;
	va_start(args, format);
  vsnprintf(buffer, sizeof(buffer) - 1, format, args);
  errs() << buffer;
	va_end(args);
}

void ModularizePPCallbacks::Trace( 
    SourceLocation Loc, const char *format, ...) {
  char buffer[1024];
	va_list	args;
	va_start(args, format);
  errs() << SourceLocationStr(Loc) << ": ";
  vsnprintf(buffer, sizeof(buffer) - 1, format, args);
  errs() << buffer;
	va_end(args);
}

/// \brief Outputs a trace string with a location from a token.
void ModularizePPCallbacks::Trace(const Token &Token, const char *format, ...) {
  char buffer[1024];
	va_list	args;
	va_start(args, format);
  errs() << SourceLocationStr(Token.getLocation()) << ": ";
  vsnprintf(buffer, sizeof(buffer) - 1, format, args);
  errs() << buffer;
	va_end(args);
}

}

