//===--- ModularizePPCallbacks.h - Callbacks for Modularize PP -*- C++ --*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
///
/// \file
/// \brief Implements the PPCallbacks interface for modularize.
///
//===----------------------------------------------------------------------===//

#ifndef MODULARIZEPPCALLBACKS_H
#define MODULARIZEPPCALLBACKS_H

#include "clang/Lex/Preprocessor.h"
#include "clang/Basic/SourceLocation.h"
#include "ModularizeHeaderFile.h"
#include "ModularizeMasterHeaderTracker.h"

namespace Modularize {
  using namespace clang;
  using namespace llvm;

typedef StringMap<std::string> ModularizeSymbolMap;
typedef ModularizeSymbolMap::iterator ModularizeSymbolMapIterator;

// FIXME: This class doesn't need to override all the handlers.

/// \brief This class provides callback implementations to grab information
///   about preprocessor directives encountered.
class ModularizePPCallbacks : public PPCallbacks {
public:
  ModularizePPCallbacks(Preprocessor &pp, StringRef topFile,
	  ModularizeMasterHeaderTracker &masterHeaderTracker);
  ~ModularizePPCallbacks();

  /// \brief Callback invoked whenever a source file is entered or exited.
  ///
  /// \param Loc Indicates the new location.
  /// \param PrevFID the file that was exited if \p Reason is ExitFile.
  void FileChanged(SourceLocation Loc, FileChangeReason Reason,
                           SrcMgr::CharacteristicKind FileType,
                           FileID PrevFID = FileID());

  /// \brief Callback invoked whenever a source file is skipped as the result
  /// of header guard optimization.
  ///
  /// \param ParentFile The file that \#included the skipped file.
  ///
  /// \param FilenameTok The token in ParentFile that indicates the
  /// skipped file.
  void FileSkipped(const FileEntry &ParentFile,
                           const Token &FilenameTok,
                           SrcMgr::CharacteristicKind FileType);

  /// \brief Callback invoked whenever an inclusion directive results in a
  /// file-not-found error.
  ///
  /// \param FileName The name of the file being included, as written in the 
  /// source code.
  ///
  /// \param RecoveryPath If this client indicates that it can recover from 
  /// this missing file, the client should set this as an additional header
  /// search patch.
  ///
  /// \returns true to indicate that the preprocessor should attempt to recover
  /// by adding \p RecoveryPath as a header search path.
  bool FileNotFound(StringRef FileName,
                            SmallVectorImpl<char> &RecoveryPath);

  /// \brief Callback invoked whenever an inclusion directive of
  /// any kind (\c \#include, \c \#import, etc.) has been processed, regardless
  /// of whether the inclusion will actually result in an inclusion.
  ///
  /// \param HashLoc The location of the '#' that starts the inclusion 
  /// directive.
  ///
  /// \param IncludeTok The token that indicates the kind of inclusion 
  /// directive, e.g., 'include' or 'import'.
  ///
  /// \param FileName The name of the file being included, as written in the 
  /// source code.
  ///
  /// \param IsAngled Whether the file name was enclosed in angle brackets;
  /// otherwise, it was enclosed in quotes.
  ///
  /// \param FilenameRange The character range of the quotes or angle brackets
  /// for the written file name.
  ///
  /// \param File The actual file that may be included by this inclusion 
  /// directive.
  ///
  /// \param SearchPath Contains the search path which was used to find the file
  /// in the file system. If the file was found via an absolute include path,
  /// SearchPath will be empty. For framework includes, the SearchPath and
  /// RelativePath will be split up. For example, if an include of "Some/Some.h"
  /// is found via the framework path
  /// "path/to/Frameworks/Some.framework/Headers/Some.h", SearchPath will be
  /// "path/to/Frameworks/Some.framework/Headers" and RelativePath will be
  /// "Some.h".
  ///
  /// \param RelativePath The path relative to SearchPath, at which the include
  /// file was found. This is equal to FileName except for framework includes.
  ///
  /// \param Imported The module, whenever an inclusion directive was
  /// automatically turned into a module import or null otherwise.
  ///
  void InclusionDirective(SourceLocation HashLoc,
                                  const Token &IncludeTok,
                                  StringRef FileName,
                                  bool IsAngled,
                                  CharSourceRange FilenameRange,
                                  const FileEntry *File,
                                  StringRef SearchPath,
                                  StringRef RelativePath,
                                  const Module *Imported);

  /// \brief Callback invoked whenever there was an explicit module-import
  /// syntax.
  ///
  /// \param ImportLoc The location of import directive token.
  ///
  /// \param Path The identifiers (and their locations) of the module
  /// "path", e.g., "std.vector" would be split into "std" and "vector".
  ///
  /// \param Imported The imported module; can be null if importing failed.
  ///
  void moduleImport(SourceLocation ImportLoc,
                            ModuleIdPath Path,
                            const Module *Imported);

  /// \brief Callback invoked when the end of the main file is reached.
  ///
  /// No subsequent callbacks will be made.
  void EndOfMainFile();

  /// \brief Callback invoked when a \#ident or \#sccs directive is read.
  /// \param Loc The location of the directive.
  /// \param str The text of the directive.
  ///
  void Ident(SourceLocation Loc, const std::string &str);

  /// \brief Callback invoked when a \#pragma comment directive is read.
  void PragmaComment(SourceLocation Loc, const IdentifierInfo *Kind,
                             const std::string &Str);

  /// \brief Callback invoked when a \#pragma clang __debug directive is read.
  /// \param Loc The location of the debug directive.
  /// \param DebugType The identifier following __debug.
  void PragmaDebug(SourceLocation Loc, StringRef DebugType);

  /// \brief Callback invoked when a \#pragma message directive is read.
  /// \param Loc The location of the message directive.
  /// \param Namespace The namespace of the message directive.
  /// \param Kind The type of the message directive.
  /// \param Str The text of the message directive.
  void PragmaMessage(SourceLocation Loc, StringRef Namespace,
                             PragmaMessageKind Kind, StringRef Str);

  /// \brief Callback invoked when a \#pragma gcc dianostic push directive
  /// is read.
  void PragmaDiagnosticPush(SourceLocation Loc,
                                    StringRef Namespace);

  /// \brief Callback invoked when a \#pragma gcc dianostic pop directive
  /// is read.
  void PragmaDiagnosticPop(SourceLocation Loc,
                                   StringRef Namespace);

  /// \brief Callback invoked when a \#pragma gcc dianostic directive is read.
  void PragmaDiagnostic(SourceLocation Loc, StringRef Namespace,
                                diag::Mapping mapping, StringRef Str);

  /// \brief Called by Preprocessor::HandleMacroExpandedIdentifier when a
  /// macro invocation is found.
  void MacroExpands(const Token &MacroNameTok, const MacroDirective *MD,
                            SourceRange Range, const MacroArgs *Args);

  /// \brief Hook called whenever a macro definition is seen.
  void MacroDefined(const Token &MacroNameTok,
                            const MacroDirective *MD);

  /// \brief Hook called whenever a macro \#undef is seen.
  ///
  /// MD is released immediately following this callback.
  void MacroUndefined(const Token &MacroNameTok,
                              const MacroDirective *MD);
  
  /// \brief Hook called whenever the 'defined' operator is seen.
  /// \param MD The MacroDirective if the name was a macro, null otherwise.
  void Defined(const Token &MacroNameTok, const MacroDirective *MD);
  
  /// \brief Hook called when a source range is skipped.
  /// \param Range The SourceRange that was skipped. The range begins at the
  /// \#if/\#else directive and ends after the \#endif/\#else directive.
  void SourceRangeSkipped(SourceRange Range);

  /// \brief Hook called whenever an \#if is seen.
  /// \param Loc the source location of the directive.
  /// \param ConditionRange The SourceRange of the expression being tested.
  ///
  // FIXME: better to pass in a list (or tree!) of Tokens.
  void If(SourceLocation Loc, SourceRange ConditionRange);

  /// \brief Hook called whenever an \#elif is seen.
  /// \param Loc the source location of the directive.
  /// \param ConditionRange The SourceRange of the expression being tested.
  /// \param IfLoc the source location of the \#if/\#ifdef/\#ifndef directive.
  // FIXME: better to pass in a list (or tree!) of Tokens.
  void Elif(SourceLocation Loc, SourceRange ConditionRange,
                    SourceLocation IfLoc);

  /// \brief Hook called whenever an \#ifdef is seen.
  /// \param Loc the source location of the directive.
  /// \param MacroNameTok Information on the token being tested.
  /// \param MD The MacroDirective if the name was a macro, null otherwise.
  void Ifdef(SourceLocation Loc, const Token &MacroNameTok,
                     const MacroDirective *MD);

  /// \brief Hook called whenever an \#ifndef is seen.
  /// \param Loc the source location of the directive.
  /// \param MacroNameTok Information on the token being tested.
  /// \param MD The MacroDirective if the name was a macro, null otherwise.
  void Ifndef(SourceLocation Loc, const Token &MacroNameTok,
                      const MacroDirective *MD);

  /// \brief Hook called whenever an \#else is seen.
  /// \param Loc the source location of the directive.
  /// \param IfLoc the source location of the \#if/\#ifdef/\#ifndef directive.
  void Else(SourceLocation Loc, SourceLocation IfLoc);

  /// \brief Hook called whenever an \#endif is seen.
  /// \param Loc the source location of the directive.
  /// \param IfLoc the source location of the \#if/\#ifdef/\#ifndef directive.
  void Endif(SourceLocation Loc, SourceLocation IfLoc);
private:
  ModularizeMasterHeaderTracker &MasterHeaderTracker;
  std::string TopFile;
  Preprocessor &PP;
  ModularizeHeaderFile *RootHeaderFile;
  ModularizeHeaderFile *CurrentHeaderFile;
  ModularizeHeaderFileMap HeaderFileMap;
  ModularizeSymbolMap Symbols;

  /// \brief Lookup header.
  ModularizeHeaderFile *GetHeaderFile(const std::string& name);
  void AddHeaderFile(ModularizeHeaderFile *headerFile);
  std::string SourceLocationStr(SourceLocation Loc);
  std::string SourceLocationFile(SourceLocation Loc);
  unsigned ModularizePPCallbacks::SourceLocationLineNumber(SourceLocation Loc);
  std::string FileName(FileID fileID);
  const char *FileChangeReasonStr(FileChangeReason reason);
  const char *CharacteristicKindStr(SrcMgr::CharacteristicKind fileType);
  // Retrieve source snippet from file image.
  std::string SourceSnippet(SourceRange sourceRange);
  // Do macro substitutions in a string.
  std::string DoMacroSubstitutions(std::string str);

  // Get symbol map.
  ModularizeSymbolMap &GetSymbols() { return Symbols; }
  // Add a symbol definition.  If the symbol already exists,
  // its value is replaced.
  void AddSymbol(std::string name, std::string value);
  // Get symbol value.  Return empty string if not defined.
  std::string GetSymbol(std::string name);
  // Get symbol value.  Return true if found.
  bool GetSymbol(std::string name, std::string &value);

  /// \brief Outputs a trace string.
  void Trace(const char *format, ...);
  /// \brief Outputs a trace string with a location.
  void Trace(SourceLocation Loc, const char *format, ...);
  /// \brief Outputs a trace string with a location from a token.
  void Trace(const Token &Token, const char *format, ...);
};

}  // end namespace clang

#endif
