//===--- ModularizePPDirective.cpp - Represents a PP directive --*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//

#include "ModularizePPDirective.h"
#include "ModularizeUtilities.h"
#include "clang/Basic/SourceManager.h"
#include "llvm/Support/raw_ostream.h"
#include <stdio.h>

#if _MSC_VER >= 1400  // VC 8.0 and later deprecate snprintf and _snprintf.
# define snprintf _snprintf_s
#elif _MSC_VER
# define snprintf _snprintf
#endif

namespace Modularize {
using namespace llvm;
using namespace clang;

ModularizePPDirective::ModularizePPDirective(
	ModularizePPDirectiveKind directiveKind, unsigned lineNumber,
    std::string unconvertedConditional, std::string convertedConditional) :
  DirectiveKind(directiveKind),
  LineNumber(lineNumber),
  UnconvertedConditional(unconvertedConditional),
  ConvertedConditional(convertedConditional) {
}

ModularizePPDirective::~ModularizePPDirective() {
}

// Return true if this directive matches another.
// (It compares the converted conditionals.)
bool ModularizePPDirective::Match(ModularizePPDirective *other) {
  if (ConvertedConditional == other->GetConvertedConditional()) {
    return true;
  }
  return false;
}

// Print this directive to a string suitable for a message.
// FIXME: The following was to avoid the apparent portability
// problems with snprintf.  Perhaps there was a better way
// with just std::string or other string utilities I don't
// know about, but the following has fewer allocations.
std::string ModularizePPDirective::PrintToString(
    int level, std::string fileName) {
  char buffer[1024];
  char lineBuffer[80];
  snprintf(lineBuffer, sizeof(lineBuffer) - 1, ":%d:1", LineNumber);
  strncpy(buffer, Indent(level).c_str(), sizeof(buffer) - 1);
  strncat(buffer, fileName.c_str(), sizeof(buffer) - 1);
  strncat(buffer, lineBuffer, sizeof(buffer) - 1);
  strncat(buffer, ": ", sizeof(buffer) - 1);
  const char *directiveString;
  switch (DirectiveKind) {
    case MPPD_If:
      directiveString = "#if ";
      break;
    case MPPD_ElIf:
      directiveString = "#elif ";
      break;
    case MPPD_IfDef:
      directiveString = "#ifdef ";
      break;
    case MPPD_IfNDef:
      directiveString = "#ifndef ";
      break;
    default:
      directiveString = "#unknown ";
      break;
  }
  strncat(buffer, directiveString, sizeof(buffer) - 1);
  strncat(buffer, UnconvertedConditional.c_str(), sizeof(buffer) - 1);
  strncat(buffer, "  (", sizeof(buffer) - 1);
  strncat(buffer, directiveString, sizeof(buffer) - 1);
  strncat(buffer, ConvertedConditional.c_str(), sizeof(buffer) - 1);
  strncat(buffer, ")\n", sizeof(buffer) - 1);
  buffer[sizeof(buffer) - 1] = '\0';
  return std::string(buffer);
}

// Display the information for debugging.
void ModularizePPDirective::Dump(int level, const char *fileName) {
  if (fileName == NULL)
    fileName = "line";
  errs() << PrintToString(level, fileName);
}

}
